# -*- coding: utf-8 -*-

# Copyright 2016-2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

import traceback
from mongomotor.fields import (StringField, IntField, BooleanField)
from toxicbuild.core.exceptions import ToxicClientException, BadJsonData
from toxicbuild.core.utils import string2datetime, LoggerMixin, now
from toxicbuild.master.build import BuildStep, Builder
from toxicbuild.master.client import get_build_client
from toxicbuild.master.signals import (build_started, build_finished,
                                       step_started, step_finished,
                                       step_output_arrived)
from toxicbuild.master.utils import OwnedDocument


class Slave(OwnedDocument, LoggerMixin):

    """ Slaves are the entities that actualy do the work
    of execute steps. The comunication to slaves is through
    the network (using :class:`toxicbuild.master.client.BuildClient`).
    The steps are actually decided by the slave.
    """
    name = StringField(required=True, unique=True)
    host = StringField(required=True)
    port = IntField(required=True)
    token = StringField(required=True)
    is_alive = BooleanField(default=False)

    meta = {
        'ordering': ['name']
    }

    @classmethod
    async def create(cls, **kwargs):
        slave = cls(**kwargs)
        await slave.save()
        return slave

    def to_dict(self, id_as_str=False):
        my_dict = {'name': self.name, 'host': self.host,
                   'port': self.port, 'token': self.token,
                   'is_alive': self.is_alive, 'id': self.id}
        if id_as_str:
            my_dict['id'] = str(self.id)
        return my_dict

    @classmethod
    async def get(cls, **kwargs):
        slave = await cls.objects.get(**kwargs)
        return slave

    async def get_client(self):
        """ Returns a :class:`toxicbuild.master.client.BuildClient` instance
        already connected to the server.
        """
        connected_client = await get_build_client(self, self.host,
                                                  self.port)
        return connected_client

    async def healthcheck(self):
        """ Check if the build server is up and running
        """
        with (await self.get_client()) as client:
            alive = await client.healthcheck()

        self.is_alive = alive
        # using yield instead of await because mongomotor's
        # save returns a tornado Future, not a asyncio Future
        await self.save()
        return self.is_alive

    async def list_builders(self, revision):
        """ List builder available in for a given revision

        :param revision: An instance of
          :class:`toxicbuild.master.repository.RepositoryRevision`
        """
        repository = await revision.repository
        repo_url = repository.url
        vcs_type = repository.vcs_type
        branch = revision.branch
        named_tree = revision.commit

        with (await self.get_client()) as client:
            builders = await client.list_builders(repo_url, vcs_type,
                                                  branch, named_tree)

        builder_instnces = []
        for bname in builders:
            builder = await Builder.get_or_create(repository=repository,
                                                  name=bname)
            builder_instnces.append(builder)

        return list(builder_instnces)

    async def build(self, build):
        """ Connects to a build server and requests a build on that server

        :param build: An instance of :class:`toxicbuild.master.build.Build`
        """

        with (await self.get_client()) as client:

            try:
                build_info = await client.build(
                    build, process_coro=self._process_info)
            except (ToxicClientException, BadJsonData):
                output = traceback.format_exc()
                build.status = 'exception'
                build.started = build.started or now()
                build.finished = build.finished or now()
                exception_step = BuildStep(output=output, started=now(),
                                           finished=now(), status='exception',
                                           command='', name='exception')
                build.steps.append(exception_step)

                await build.update()
                build_info = build.to_dict()

        return build_info

    async def _process_info(self, build, info):
        """ Method used to process information sent by
        the build server about an in progress build.

        :param build: The build that is being executed
        :param info: A dictionary. The information sent by the
          slave that is executing the build.
        """

        # if we need one more conditional here is better to use
        # a map...
        if info['info_type'] == 'build_info':
            await self._process_build_info(build, info)

        elif info['info_type'] == 'step_info':
            await self._process_step_info(build, info)

        else:
            await self._process_step_output_info(build, info)

    async def _process_build_info(self, build, build_info):
        repo = await build.repository
        build.status = build_info['status']
        build.started = string2datetime(build_info['started'])
        finished = build_info['finished']
        if finished:
            build.finished = string2datetime(finished)
            build.total_time = (build.finished - build.started).seconds

        await build.update()

        if not build.finished:
            msg = 'build started at {}'.format(build_info['started'])
            self.log(msg)
            build_started.send(str(repo.id), build=build)
        else:
            msg = 'build finished at {} with status {}'.format(
                build_info['finished'], build.status)
            self.log(msg)
            build_finished.send(str(repo.id), build=build)

    async def _process_step_info(self, build, step_info):

        cmd = step_info['cmd']
        name = step_info['name']
        status = step_info['status']
        output = step_info['output']
        started = step_info['started']
        finished = step_info['finished']
        index = step_info['index']
        uuid = step_info['uuid']

        repo = await build.repository
        requested_step = self._get_step(build, uuid)

        if requested_step:
            requested_step.status = status
            requested_step.output = output
            requested_step.finished = string2datetime(finished)
            requested_step.total_time = step_info['total_time']
            msg = 'step {} finished at {} with status {}'.format(
                requested_step.command, finished, requested_step.status)
            self.log(msg, level='debug')
            step_finished.send(str(repo.id), build=build, step=requested_step)

        else:
            requested_step = BuildStep(name=name, command=cmd,
                                       status=status, output=output,
                                       started=string2datetime(started),
                                       index=index, uuid=uuid)
            msg = 'step {} started at {}'.format(requested_step.command,
                                                 started)
            self.log(msg, level='debug')
            step_started.send(str(repo.id), build=build, step=requested_step)
            build.steps.append(requested_step)

        await build.update()

    async def _process_step_output_info(self, build, info):
        uuid = info['uuid']
        info['output'] = info['output'] + '\n'
        output = info['output']
        repo = await build.repository
        step = self._get_step(build, uuid)
        step.output = ''.join([step.output or '', output])
        info['repository'] = {'id': str(repo.id)}
        await build.update()
        msg = 'step_output_arrived for {}'.format(uuid)
        self.log(msg, level='debug')
        step_output_arrived.send(str(repo.id), step_info=info)

    def _get_step(self, build, step_uuid):
        """Returns a step from ``build``. Returns None if the requested
        step is not present in the build.

        :param build: A :class:`toxicbuild.master.build.Build` instance.
        :param step_uuid: The uuid of the requested step.
        """

        for step in build.steps:
            if step.uuid == step_uuid:
                return step



# -*- coding: utf-8 -*-

# This module contains no utilities. No matter what you are looking
# for, it is not here.

# flake8: noqa

SENTENCES = [
    'A metafísica é uma consequência de estar mal disposto.',
    'Se os tubarões fossem homens, tratariam melhor dos peixes pequenos.',
    'Eu verifico que não tenho par nisto tudo neste mundo.',
    'O ódio contra a baixeza também endurece os rostos.',
    'Ah, que maçada quererem que eu seja de companhia!',
    'Vivemos num tempo de chantagem universal',
    'Queria casar, conseguiu uma submulher. Tiveram subfilhos. Subnutridos.',
    ' A gente se acostuma para poupar a vida.',
    'No dia em que a Muralha da China ficou pronta, para onde foram os pedreiros?'
]


LOGOS = [
    [
        ('bold_blue', ' _____         _\n'),
        ('bold_blue', '|_   _|       (_)\n'),
        ('bold_blue', '  | | ___'), ('bold_yellow', '__  __'), ('bold_blue',
                                                                '_  ___ '), ('bold_red', ' ___      '), ('bold_white', '_'), ('bold_red', ' _    _\n'),
        ('bold_blue', '  | |/ _ '), ('bold_yellow', '\ \/ /'), ('bold_blue',
                                                                ' |/ __|'), ('bold_red', '| _ )_  _'), ('bold_white', '(_)'), ('bold_red', ' |__| |\n'),
        ('bold_blue', '  | | (_)'), ('bold_yellow', ' >  <'), ('bold_blue',
                                                               '| | (__ '), ('bold_red', '| _ \ || '), ('bold_white', '| |'), ('bold_red', ' / _` |\n'),
        ('bold_blue', '  \_/\___'), ('bold_yellow', '/_/\_\\'), ('bold_blue',
                                                                 '_|\___|'), ('bold_red', '|___/\_,_'), ('bold_white', '|_|'), ('bold_red', '_\__,_|\n')
    ],

    [
        ('bold_blue', '  __\n'),
        ('bold_blue',  ' /|            /\n'),
        ('bold_blue',
         '( |  ___         ___'), ('bold_red', '  __              ___\n'),
        ('bold_blue', '  | |   )'), ('bold_yellow', '(_/_'), ('bold_blue', ' | |'), ('bold_red',
                                                                                     '    |__] |  |'), ('bold_white', ' |'), ('bold_red', ' |    |  \ \n'),
        ('bold_blue', '  | |__/ '), ('bold_yellow', ' / /'), ('bold_blue', ' | |__'), ('bold_red',
                                                                                       '  |__] |__|'), ('bold_white', ' |'), ('bold_red', ' |___ |__/\n'),

    ],

    [
        ('bold_blue', '___________          .__        '), ('bold_red',
                                                            "'||                "), ('bold_white', '||'), ('bold_red', "  '||       '||\n"),
        ('bold_blue', '\__    ___/___'), ('bold_yellow', '___  __'), ('bold_blue', '|__| ____   '), ('bold_red',
                                                                                                     "|| ...  ... ...  "), ('bold_white', "..."), ("bold_red", "   ||     .. ||\n"),
        ('bold_blue', '  |    | /  _ '), ('bold_yellow', '\  \/  /'), ('bold_blue', '  |/ ___\  '), ('bold_red',
                                                                                                     "||'  ||  ||  ||   "), ('bold_white', "||"), ('bold_red', "   ||   .'  '||\n"),
        ('bold_blue', '  |    |(  <_> '), ('bold_yellow', '>    <'), ('bold_blue', '|  \  \___  '), (
            'bold_red', "||    |  ||  ||   "), ('bold_white', "||"), ('bold_red', "   ||   |.   ||\n"),
        ('bold_blue', '  |____| \____'), ('bold_yellow', '/__/\_ \\'), ('bold_blue', '__|\___  > '), ('bold_red',
                                                                                                      "'|...'   '|..'|. "), ('bold_white', ".||."), ('bold_red', " .||.  '|..'||.\n"),
        ('bold_blue', '                    '), ('bold_yellow',
                                                '\/'), ('bold_blue', '       \/')

    ]
]



# -*- coding: utf-8 -*-

from toxicbuild.core.vcs import get_vcs
from toxicbuild.core.protocol import BaseToxicProtocol
from toxicbuild.core.client import BaseToxicClient


make_pyflakes_happy = [get_vcs, BaseToxicProtocol, BaseToxicClient]

del make_pyflakes_happy



# -*- coding: utf-8 -*-

# Copyright 2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

from mongomotor import Document, EmbeddedDocument
from mongomotor.fields import (StringField, UUIDField, ListField,
                               ReferenceField, EmbeddedDocumentListField,
                               BooleanField, EmailField)
from mongomotor.queryset import PULL
from toxicbuild.core.utils import bcrypt_string
from toxicbuild.master import settings
from toxicbuild.master.exceptions import InvalidCredentials
from toxicbuild.master.utils import as_db_ref


class Organization(Document):
    uuid = UUIDField()

    name = StringField(required=True, unique=True)
    owner = ReferenceField('User', required=True)
    teams = EmbeddedDocumentListField('Team')

    async def add_user(self, user):
        """Adds a user to the organization"""

        await user.update(push__member_of=self)

    async def remove_user(self, user):
        """Removes a user from the organization."""

        await user.update(pull__member_of=self)

    @property
    def users(self):
        return User.objects.filter(member_of=self)

    async def save(self, *args, **kwargs):
        await super().save(*args, **kwargs)
        # we set it here so we can query for user's repo in a easier way
        owner = await self.owner
        # do not deref to avoid a query for all orgs
        organizations = [ref.id for ref in as_db_ref(owner, 'organizations')]
        if self.id not in organizations:
            await owner.update(push__organizations=self)


class User(Document):
    uuid = UUIDField()

    email = EmailField(required=True, unique=True)
    username = StringField(required=True, unique=True)
    password = StringField()
    is_superuser = BooleanField(default=False)
    # organizations owned by the user
    organizations = ListField(ReferenceField('Organization',
                                             reverse_delete_rule=PULL))
    # organizations which the user is part of, but not the owner
    member_of = ListField(ReferenceField('Organization'))
    # what the user can do: create_repo, create_slave or create_user
    allowed_actions = ListField(StringField())

    @classmethod
    async def authenticate(cls, username_or_email, password):
        """Authenticates an user. Returns an user if the user is
        authenticated. Raises ``InvalidCredentials`` if a user with
        this credentials does not exist.

        :param username_or_email: Username or email to use to authenticate.
        :param password: Not encrypted password."""

        salt = settings.BCRYPT_SALT
        password = bcrypt_string(password, salt)
        fields = ['username', 'email']
        for field in fields:
            kw = {field: username_or_email, 'password': password}
            try:
                user = await cls.objects.get(**kw)
                return user
            except cls.DoesNotExist:
                pass

        raise InvalidCredentials

    def to_dict(self):
        objdict = {'id': str(self.id),
                   'username': self.username,
                   'email': self.email}
        return objdict

    async def save(self, *args, **kwargs):
        if not self.username:
            self.username = self.email.split('@')[0]

        r = await super().save(*args, **kwargs)
        return r

    async def delete(self, *args, **kwargs):
        r = await super().delete(*args, **kwargs)
        await Organization.objects.filter(owner=self).delete()
        return r

    def set_password(self, password):
        salt = settings.BCRYPT_SALT
        self.password = bcrypt_string(password, salt)


class Team(EmbeddedDocument):
    uuid = UUIDField()

    name = StringField(required=True)
    members = ListField(ReferenceField(User))



# -*- coding: utf-8 -*-

# Copyright 2015 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

from asyncio import ensure_future
from toxicbuild.core import BaseToxicClient


class BuildClient(BaseToxicClient):

    """ A client to :class:`toxicbuild.slave.server.BuildServer`
    """

    def __init__(self, slave, *args, **kwargs):
        self.slave = slave
        super().__init__(*args, **kwargs)

    async def healthcheck(self):
        """ Asks to know if the server is up and running
        """
        data = {'action': 'healthcheck'}
        try:
            await self.write(data)
            response = await self.get_response()
            del response
            return True
        except Exception:
            return False

    async def list_builders(self, repo_url, vcs_type, branch, named_tree):
        """ Asks the server for the builders available for ``repo_url``,
        on ``branch`` and ``named_tree``.
        """

        data = {'action': 'list_builders',
                'body': {'repo_url': repo_url,
                         'vcs_type': vcs_type,
                         'branch': branch,
                         'named_tree': named_tree}}
        await self.write(data)
        response = await self.get_response()
        builders = response['body']['builders']
        return builders

    async def build(self, build, process_coro=None):
        """Requests a build for the build server.

        :param build: The build that will be executed.
        :param process_coro: A coroutine to process the intermediate
          build information sent by the build server."""

        repository = await build.repository
        builder_name = (await build.builder).name
        slave = await build.slave
        data = {'action': 'build',
                'token': slave.token,
                'body': {'repo_url': repository.url,
                         'vcs_type': repository.vcs_type,
                         'branch': build.branch,
                         'named_tree': build.named_tree,
                         'builder_name': builder_name}}

        await self.write(data)
        futures = []
        while True:
            r = await self.get_response()

            if not r:
                break

            build_info = r['body']
            if process_coro:
                future = ensure_future(process_coro(build, build_info))
                futures.append(future)

        return futures


async def get_build_client(slave, addr, port):
    """ Instanciate :class:`toxicbuild.master.client.BuildClient` and
    connects it to a build server
    """

    client = BuildClient(slave, addr, port)
    await client.connect()
    return client



# -*- coding: utf-8 -*-

import gettext
import os
import pkg_resources
import shutil
import sys
from tornado.platform.asyncio import AsyncIOMainLoop
from pyrocumulus.commands.base import get_command
from toxicbuild.core.cmd import command, main
from toxicbuild.core.conf import Settings
from toxicbuild.core.utils import bcrypt, changedir


here = os.path.dirname(os.path.abspath(__file__))
translations = os.path.join(here, 'translations')

gettext.install('toxicbuild.ui', translations)

ENVVAR = 'TOXICUI_SETTINGS'
DEFAULT_SETTINGS = 'toxicui.conf'

settings = None

LOGFILE = './toxicui.log'


def create_settings():
    global settings

    settings = Settings(ENVVAR, DEFAULT_SETTINGS)


# we install the asyncio loop for tornado so we can
# use it inside tornado handlers with @gen.coroutine
AsyncIOMainLoop().install()


pyrocommand = None


def _check_conffile(workdir, conffile):
    """Checks if the conffile is inside workdir."""

    absworkdir = os.path.abspath(workdir)
    absconffile = os.path.abspath(conffile)

    return absconffile.startswith(absworkdir)


@command
def start(workdir, daemonize=False, stdout=LOGFILE, stderr=LOGFILE,
          pidfile=None, loglevel='info', conffile=None):
    """ Starts the web interface.

    Starts the build server to listen on the specified port for
    requests from addr (0.0.0.0 means everyone). Addr and port params
    came from the configfile

    :param workdir: Work directory for server.
    :param --daemonize: Run as daemon. Defaults to False
    :param --stdout: stdout path. Defaults to /dev/null
    :param --stderr: stderr path. Defaults to /dev/null
    :param --pidfile: pid file for the process.
    :param --loglevel: Level for logging messages. Defaults to `info`.
    :param -c, --conffile: path to config file. It must be relative
      to the workdir. Defaults to None. If not conffile, will look
      for a file called ``toxicui.conf`` inside ``workdir``
    """

    global pyrocommand

    if not os.path.exists(workdir):
        print('Workdir `{}` does not exist'.format(workdir))
        sys.exit(1)

    workdir = os.path.abspath(workdir)
    with changedir(workdir):
        sys.path.append(workdir)

        if conffile:

            is_in_workdir = _check_conffile(workdir, conffile)

            if not is_in_workdir:
                print('Config file must be inside workdir')
                sys.exit(1)

            os.environ['TOXICUI_SETTINGS'] = os.path.join(workdir, conffile)
            module = conffile.replace('.conf', '').replace(
                workdir, '').strip('/').replace(os.sep, '.')
            os.environ['PYROCUMULUS_SETTINGS_MODULE'] = module
        else:
            os.environ['TOXICUI_SETTINGS'] = os.path.join(workdir,
                                                          'toxicui.conf')
            os.environ['PYROCUMULUS_SETTINGS_MODULE'] = 'toxicui'

        create_settings()

        sys.argv = ['pyromanager.py', '']

        if not pyrocommand:
            pyrocommand = command = get_command('runtornado')()
        else:
            command = pyrocommand

        command.kill = False
        command.daemonize = daemonize
        command.stderr = stderr
        command.application = None
        command.stdout = stdout
        command.port = settings.TORNADO_PORT
        command.pidfile = pidfile
        command.run()


@command
def stop(workdir, pidfile=None):
    """ Stops the web interface.

    :param workdir: Work directory for the ui to be killed.
    :param --pidfile: pid file for the process.
    """

    global pyrocommand

    if not os.path.exists(workdir):
        print('Workdir `{}` does not exist'.format(workdir))
        sys.exit(1)

    workdir = os.path.abspath(workdir)
    with changedir(workdir):
        sys.path.append(workdir)

        os.environ['TOXICUI_SETTINGS'] = os.path.join(workdir,
                                                      'toxicui.conf')
        os.environ['PYROCUMULUS_SETTINGS_MODULE'] = 'toxicui'

        create_settings()

        sys.argv = ['pyromanager.py', '']

        if not pyrocommand:
            pyrocommand = command = get_command('runtornado')()
        else:
            command = pyrocommand

        command.pidfile = pidfile
        command.kill = True
        command.run()


@command
def restart(workdir, pidfile=None):
    """Restarts toxicslave

    The instance of toxicweb in ``workdir`` will be restarted.
    :param workdir: Workdir for master to be killed.
    :param --pidfile: Name of the file to use as pidfile.
    """

    stop(workdir, pidfile=pidfile)
    start(workdir, pidfile=pidfile, daemonize=True)


@command
def create(root_dir, access_token):
    """ Create a new toxicweb project.

    :param --root_dir: Root directory for toxicweb.
    :param --access-token: Access token to master's hole.
    """
    print('Creating root_dir {}'.format(root_dir))

    os.makedirs(root_dir)

    template_fname = 'toxicui.conf.tmpl'
    template_dir = pkg_resources.resource_filename('toxicbuild.ui',
                                                   'templates')
    template_file = os.path.join(template_dir, template_fname)
    dest_file = os.path.join(root_dir, 'toxicui.conf')
    shutil.copyfile(template_file, dest_file)

    # cookie secret to tornado secure cookies
    cookie_secret = bcrypt.gensalt(8).decode()

    # and finally update the config file content with the new generated
    # salt and access token
    with open(dest_file, 'r+') as fd:
        content = fd.read()
        content = content.replace('{{HOLE_TOKEN}}', access_token)
        content = content.replace('{{COOKIE_SECRET}}', cookie_secret)
        fd.seek(0)
        fd.write(content)

    print('Toxicui environment created for web ')
    return access_token


def _ask_thing(thing):
    response = input(thing)
    while not response:
        response = input(thing)

    return response


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright 2015, 2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

import asyncio
try:
    from asyncio import ensure_future
except ImportError:  # pragma: no cover
    from asyncio import async as ensure_future
from collections import OrderedDict
import json
import time
import traceback
from toxicbuild.core import utils


class BaseToxicProtocol(asyncio.StreamReaderProtocol, utils.LoggerMixin):

    """ Base protocol for toxicbulid servers
    """

    # Salt used to encrypt incomming token.
    salt = None
    # This is the token used to authenticate incomming requests.
    encrypted_token = None

    def __init__(self, loop, connection_lost_cb=None):
        """:param loop: An asyncio loop.
        :param connection_lost_cb: Callable to be executed when connection
          is closed. Gets an exception or None as parameter."""

        self.raw_data = None
        self.json_data = None
        self.action = None
        self._connected = False
        # make tests easier
        self._check_data_future = None
        self._client_connected_future = None
        self.connection_lost_cb = connection_lost_cb

        reader = asyncio.StreamReader(loop=loop)
        super().__init__(reader, loop=loop)

    def __call__(self):
        return self

    def connection_made(self, transport):
        """ Called once, when the client connects.

        :param transport: transport for asyncio.StreamReader and
          asyncio.StreamWriter.
        """
        self._transport = transport
        self._stream_reader.set_transport(transport)
        self._stream_writer = asyncio.StreamWriter(transport, self,
                                                   self._stream_reader,
                                                   self._loop)
        self._connected = True

        self.peername = self._transport.get_extra_info('peername')
        self._check_data_future = ensure_future(self.check_data())
        self._check_data_future.add_done_callback(self._check_data_cb)

    def connection_lost(self, exc):
        self.close_connection()
        super().connection_lost(exc)
        if self.connection_lost_cb:
            self.connection_lost_cb(exc)

    @asyncio.coroutine
    def check_data(self):
        """ Checks if the data is valid, it means, checks if has some data,
        checks if it is a valid json and checks if it has a ``action`` key
        """

        self.data = yield from self.get_json_data()

        if not self.data:
            msg = 'Something wrong with your data {!r}'.format(self.raw_data)
            yield from self.send_response(code=1, body={'error': msg})
            return self.close_connection()

        token = self.data.get('token')
        if not token:
            msg = 'No auth token'
            yield from self.send_response(code=2, body={'error': msg})
            return self.close_connection()

        incomming_token = utils.bcrypt_string(token, self.salt)
        if incomming_token != self.encrypted_token:
            msg = 'Bad auth token'
            yield from self.send_response(code=3, body={'error': msg})
            return self.close_connection()

        self.action = self.data.get('action')

        if not self.action:
            msg = 'No action found!'
            yield from self.send_response(code=1, body=msg)
            return self.close_connection()

    @asyncio.coroutine
    def client_connected(self):  # pragma no cover
        """ Coroutine that handles connections. You must implement this
        in your sub-classes. When this method is called, ``self.data``,
        containing a dictionary with the data passed in the request and
        ``self.action``, a string indicating which action to take are already
        available.
        """
        raise NotImplementedError

    def close_connection(self):
        """ Closes the connection with the client
        """

        self._stream_writer.close()
        self._connected = False

    @asyncio.coroutine
    def send_response(self, code, body):
        """ Send a response to client formated by the (unknown) toxicbuild
        remote build specs.
        :param code: code for this message. code == 0 is success and
          code > 0 is error.
        :param body: response body. It has to be a serializable object.
        """
        response = OrderedDict()
        response['code'] = code
        response['body'] = body
        data = json.dumps(response)
        yield from utils.write_stream(self._stream_writer, data)

    @asyncio.coroutine
    def get_raw_data(self):
        """ Returns the raw data sent by the client
        """
        return utils.read_stream(self._stream_reader)

    @asyncio.coroutine
    def get_json_data(self):
        """Returns the json sent by the client."""

        data = (yield from self.get_raw_data()).decode()

        try:
            data = json.loads(data)
        except Exception:
            msg = '{}\n{}'.format(traceback.format_exc(), data)
            self.log(msg, level='error')
            data = None

        return data

    def _check_data_cb(self, future):
        # The thing here is: run client_connected only if everything ok
        # on check_data
        if not self._connected:  # pragma no cover
            return

        # wrapping it to log it.
        @asyncio.coroutine
        def logged_cb():
            init = (time.time() * 1e3)
            try:
                status = yield from self.client_connected()
            except ConnectionResetError:
                status = 1
                msg = 'Connection reset'
                self.log(msg, level='debug')

            end = (time.time() * 1e3)
            self.log('{}: {} {}'.format(self.action, status, (end - init)))

        self._client_connected_future = logged_cb()
        self._client_connected_future = ensure_future(
            self._client_connected_future)



# -*- coding: utf-8 -*-

# Copyright 2016 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

from functools import partial
import sys
from mando import Program


class ToxicProgram(Program):

    """Transforms a python function into a command line program.
    Extends mando's Program to execute _generate_command() only
    when execute() is called to enable toxicbuild command hacks."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._generate_queue = []

    def _generate_command(self, *args, **kwargs):
        super_gen = super()._generate_command
        call = partial(super_gen, *args, **kwargs)
        self._generate_queue.append(call)

        if len(args) == 1 and hasattr(args[0], '__call__'):
            return args[0]
        else:
            def _command(func):  # pragma no cover
                return func
            return _command

    def execute(self, *args, **kwargs):
        [call() for call in self._generate_queue]
        super().execute(*args, **kwargs)
        self._generate_queue = []

    def __call__(self):
        cmds = ['start', 'stop', 'create', 'restart']

        if len(sys.argv) == 1 or (len(sys.argv) == 2 and sys.argv[1] in cmds):
            sys.argv.append('-h')

        super().__call__()


main = ToxicProgram()
command = main.command
arg = main.arg
parse = main.parse
execute = main.execute



# -*- coding: utf-8 -*-

# Copyright 2015 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

from asyncblink import signal


revision_added = signal('revision-added')
step_started = signal('step-started')
step_finished = signal('step-finished')
step_output_arrived = signal('step-output-arrived')
build_started = signal('build-started')
build_finished = signal('build-finished')
build_added = signal('build-added')
repo_status_changed = signal('repo-status-changed')
repo_added = signal('repo-added')



# -*- coding: utf-8 -*-

# Copyright 2015-2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

import asyncio
from toxicbuild.core import BaseToxicClient
from toxicbuild.core.exceptions import ToxicClientException
from toxicbuild.ui import settings
from toxicbuild.ui.exceptions import UserDoesNotExist, NotEnoughPerms


class UIHoleClient(BaseToxicClient):

    def __init__(self, requester, *args, hole_token=None):
        self.hole_token = hole_token or settings.HOLE_TOKEN
        self.requester = requester
        super().__init__(*args)

    def __getattr__(self, name):
        action = name.replace('_', '-')

        @asyncio.coroutine
        def _2serverandback(**kwargs):
            return (yield from self.request2server(action, body=kwargs))

        return _2serverandback

    @asyncio.coroutine
    def request2server(self, action, body):

        data = {'action': action, 'body': body,
                'token': self.hole_token}

        if action != 'user-authenticate':
            data['user_id'] = str(self.requester.id)

        yield from self.write(data)
        response = yield from self.get_response()
        return response['body'][action]

    @asyncio.coroutine
    def connect2stream(self):
        """Connects the client to the master's hole stream."""

        action = 'stream'
        body = {'user_id': str(self.requester.id)}

        yield from self.request2server(action, body)

    @asyncio.coroutine
    def get_response(self):
        response = yield from self.read()

        if 'code' in response and int(response['code']) == 1:
            # server error
            raise ToxicClientException(response['body']['error'])

        if 'code' in response and int(response['code']) == 2:
            raise UserDoesNotExist(response['body']['error'])

        if 'code' in response and int(response['code']) == 3:
            raise NotEnoughPerms(response['body']['error'])

        return response


@asyncio.coroutine
def get_hole_client(requester, host, port, hole_token=None):
    client = UIHoleClient(requester, host, port, hole_token=hole_token)
    yield from client.connect()
    return client



# -*- coding: utf-8 -*-

# Copyright 2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

from bson.dbref import DBRef
from mongoengine.queryset import Q
from mongomotor import Document
from mongomotor.fields import GenericReferenceField
from toxicbuild.master.exceptions import NotEnoughPerms


def as_db_ref(document, field):
    """Returns reference field of a document as DBRefs."""

    try:
        dref = document._fields[field]._auto_dereference
        document._fields[field]._auto_dereference = False
        dbref = getattr(document, field)
    finally:
        document._fields[field]._auto_dereference = dref

    return dbref


class OwnedDocument(Document):

    owner = GenericReferenceField(required=True)

    meta = {'abstract': True}

    @classmethod
    async def get_for_user(cls, user, **kwargs):
        """Returns a repository if ``user`` has permission for it.
        If not raises an error.

        :param user: User who is requesting the repository.
        :param kwargs: kwargs to match the repository.
        """
        obj = await cls.objects.get(**kwargs)
        has_perms = await obj._check_perms(user)
        if not has_perms:
            msg = 'The user {} has no permissions for this object'.format(
                user.id)
            raise NotEnoughPerms(msg)
        return obj

    @classmethod
    def list_for_user(cls, user, **kwargs):
        """Returns a queryset of repositories in which the user has read
        permission"""
        member_of = [ref.id for ref in as_db_ref(user, 'member_of')]
        organizations = [ref.id for ref in as_db_ref(user, 'organizations')]

        if user.is_superuser:
            qs = cls.objects

        else:
            qs = cls.objects(
                Q(owner=user) |
                Q(__raw__={'owner._ref.$id': {'$in': organizations}}) |
                Q(__raw__={'owner._ref.$id': {'$in': member_of}}))

        return qs.filter(**kwargs)

    async def _check_perms(self, user):
        if user.is_superuser:
            return True

        owner = await self.owner
        owner_id = DBRef(owner.__class__.__name__.lower(), owner.id)
        member_of = as_db_ref(user, 'member_of')
        organizations = as_db_ref(user, 'organizations')
        if owner == user or owner_id in member_of or \
           owner_id in organizations:
            has_perms = True
        else:
            has_perms = False

        return has_perms



# -*- coding: utf-8 -*-

# Copyright 2015, 2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

import asyncio
from collections import OrderedDict
import json
import traceback
from toxicbuild.core import utils
from toxicbuild.core.exceptions import ToxicClientException, BadJsonData


class BaseToxicClient(utils.LoggerMixin):

    """ Base client for access toxicbuild servers. """

    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.loop = asyncio.get_event_loop()
        self.reader = None
        self.writer = None
        self._connected = False

    def __enter__(self):
        if not self._connected:
            msg = 'You must connect with "yield from client.connect()" '
            msg += 'before you can __enter__ on it. Sorry.'
            raise ToxicClientException(msg)
        return self

    async def __aenter__(self):
        await self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.__exit__(exc_type, exc_val, exc_tb)

    @asyncio.coroutine
    def connect(self):
        self.reader, self.writer = yield from asyncio.open_connection(
            self.host, self.port, loop=self.loop)
        self._connected = True

    def disconnect(self):
        self.log('disconecting...', level='debug')
        self.writer.close()
        self._connected = False

    @asyncio.coroutine
    def write(self, data):
        """ Writes ``data`` to the server.

        :param data: Data to be sent to the server. Will be converted to
          json and enconded using utf-8.
        """
        data = json.dumps(data)
        yield from utils.write_stream(self.writer, data)

    @asyncio.coroutine
    def read(self):
        # '{}' is decoded as an empty dict, so in json
        # context we can consider it as being a False json
        data = yield from utils.read_stream(self.reader)
        data = data.decode() or '{}'

        try:
            json_data = json.loads(data, object_pairs_hook=OrderedDict)
        except Exception:
            msg = traceback.format_exc()
            self.log(msg, level='error')
            raise BadJsonData(data)

        return json_data

    @asyncio.coroutine
    def get_response(self):
        response = yield from self.read()

        if 'code' in response and int(response['code']) != 0:
            raise ToxicClientException(response['body']['error'])
        return response



# -*- coding: utf-8 -*-


class UIFunctionNotFound(Exception):
    pass


class CloneException(Exception):
    pass


class DBError(Exception):
    pass


class NotEnoughPerms(Exception):
    pass


class OwnerDoesNotExist(Exception):
    pass


class InvalidCredentials(Exception):
    pass



# -*- coding: utf-8 -*-

# Copyright 2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

import traceback
from toxicbuild.core.client import BaseToxicClient
from toxicbuild.core.utils import LoggerMixin
from toxicbuild.slave import settings
from toxicbuild.slave.exceptions import NotConnected


class ContainerBuildClient(LoggerMixin):
    """Client to request builds to the build containers."""

    def __init__(self, build_data, container):
        """:param build_data: Data from the build request that will
        be sent to the slave in a container.
        :param container: A instance of a container that executes builds."""
        self.build_data = build_data
        self.container = container
        self.client = None

    async def __aenter__(self):
        ip = await self.container.get_container_ip()
        port = settings.CONTAINER_SLAVE_PORT

        self.client = BaseToxicClient(ip, port)
        await self.client.__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.client.__aexit__(exc_type, exc_val, exc_tb)
        self.client = None

    async def build(self, outfn):
        """Request a build to a slave in a container

        :param outfn: A callable that sends the build information back to
          the original requester."""

        token = settings. CONTAINER_SLAVE_TOKEN
        self.build_data['token'] = token

        if not self.client:
            raise NotConnected('You must __aenter__ before using it.')

        await self.client.write(self.build_data)

        while True:
            r = await self.client.get_response()

            if not r:
                break

            await outfn(r['code'], r['body'])

    async def healthcheck(self):
        token = settings. CONTAINER_SLAVE_TOKEN

        if not self.client:
            raise NotConnected('You must __aenter__ before using it.')

        data = {'action': 'healthcheck',
                'token': token,
                'body': {}}

        await self.client.write(data)

        try:
            r = await self.client.get_response()
        except Exception:
            msg = traceback.format_exc()
            self.log(msg, level='debug')
            ok = False
        else:
            ok = int(r['code']) == 0
        return ok



# -*- coding: utf-8 -*-

# Copyright 2015 2016 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

import traceback
from toxicbuild.core.vcs import get_vcs
from toxicbuild.core.utils import LoggerMixin, MatchKeysDict
from toxicbuild.master.exceptions import CloneException
from toxicbuild.master.signals import revision_added


class Poller(LoggerMixin):

    """ Class to poll changes from a vcs, process them and notificate about
    incoming changes
    """

    def __init__(self, repository, vcs_type, workdir):
        """:param repository: An instance of
          :class:`toxicbuild.repositories.Repository`

        :param vcs_type: Vcs type for :func:`toxicbuild.core.vcs.get_vcs`.
        :param workdir: workdir for vcs.
        """
        self.repository = repository
        self.vcs = get_vcs(vcs_type)(workdir)
        self._is_polling = False

    def is_polling(self):
        return self._is_polling

    async def poll(self):
        """ Check for changes on repository and if there are changes, notify
        about it.
        """

        with_clone = False
        try:
            if self.is_polling():
                self.log('alreay polling. leaving...'.format(
                    self.repository.url), level='debug')
                return

            self._is_polling = True
            self.log('Polling changes')

            if not self.vcs.workdir_exists():
                self.log('clonning repo')
                try:
                    await self.vcs.clone(self.repository.url)
                    with_clone = True
                except Exception as e:
                    msg = traceback.format_exc()
                    self.log(msg, level='error')
                    raise CloneException(str(e))

            # for git.
            # remove no branch when hg is implemented
            if hasattr(self.vcs, 'update_submodule'):  # pragma no branch
                self.log('updating submodule', level='debug')
                await self.vcs.update_submodule()

            try:
                await self.process_changes()
            except Exception as e:
                # shit happends
                msg = traceback.format_exc()
                self.log(msg, level='error')
                # but the show must go on
        finally:
            self._is_polling = False

        return with_clone

    async def process_changes(self):
        """ Process all changes since the last revision in db
        """
        self.log('processing changes', level='debug')

        dbrevisions = await self.repository.get_latest_revisions()

        since = dict((branch, r.commit_date) for branch, r
                     in dbrevisions.items() if r)

        repo_branches = MatchKeysDict(
            **{b.name: b for b in self.repository.branches})
        newer_revisions = await self.vcs.get_revisions(
            since=since, branches=repo_branches.keys())

        known_branches = dbrevisions.keys()

        revisions = []
        for branch, revs in newer_revisions.items():
            self.log('processing changes for branch {}'.format(branch),
                     level='debug')
            # the thing here is that if the branch is a new one
            # or is the first time its running, I don't what to get all
            # revisions, but the last one only.
            if branch not in known_branches:
                rev = revs[-1]
                revision = await self.repository.add_revision(branch,
                                                              **rev)
                msg = 'Last revision added for branch {} '
                self.log(msg.format(branch), level='debug')
                revisions.append(revision)
                continue

            notify_only_latest = repo_branches.get(branch).notify_only_latest \
                if repo_branches.get(branch) else False

            await self._process_branch_revisions(branch, revs,
                                                 notify_only_latest,
                                                 revisions)

        self.notify_change(*revisions)

    def notify_change(self, *revisions):
        """ Notify about new revisions added to the repository.

        :param revisions: A list of new revisions"""

        # returning for testing purposes
        return revision_added.send(self.repository, revisions=revisions)

    def log(self, msg, level='info'):
        msg = '[{}] {}'.format(self.repository.name, msg)
        super().log(msg, level)

    async def _process_branch_revisions(self, branch, revisions,
                                        notify_only_latest, to_notify):
        """Processes the revisions for a branch"""

        branch_revs = []
        for rev in revisions:
            revision = await self.repository.add_revision(branch, **rev)
            # the thing here is: if notify_only_latest, we only
            # add the most recent revision, the last one of the revisions
            # list to the revisionset
            if (not revisions[-1] == rev and notify_only_latest):
                continue

            to_notify.append(revision)
            # branch_revs just for logging
            branch_revs.append(revision)

        if branch_revs:
            msg = '{} new revisions for {} on branch {} added'
            self.log(msg.format(len(branch_revs), self.repository.url,
                                branch))



# -*- coding: utf-8 -*-

# Copyright 2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

import re
import pytz
from toxicbuild.core.conf import ConfigError
from toxicbuild.core.utils import DTFORMAT, datetime2string
from toxicbuild.ui import settings


_dt_regex = re.compile('\w+\s\w+\s\d+\s\d+:\d+:\d+\s\d+\s[\+|-]\d+$')


def _get_dtformat():
    try:
        dtformat = settings.DTFORMAT
    except ConfigError:
        dtformat = DTFORMAT

    return dtformat


def _get_timezone():
    try:
        tz = settings.TIMEZONE
        tz = pytz.timezone(tz)
    except (ConfigError, pytz.UnknownTimeZoneError):
        tz = None

    return tz


def format_datetime(dt):
    """Formats a datetime object according to the
    timezone and format specified in the config file.

    :param dt: A datetime object."""

    dtformat = _get_dtformat()
    tz = _get_timezone()

    if tz:
        dt = dt.astimezone(tz)

    return datetime2string(dt, dtformat=dtformat)


def is_datetime(dtstr):
    """Checks if a string is a formated datetime.
    The format expected for the datetime string is:
    '%a %b %d %H:%M:%S %Y %z'"""

    if not isinstance(dtstr, str):
        return False
    return bool(re.match(_dt_regex, dtstr))



# -*- coding: utf-8 -*-

# Copyright 2015-2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

from asyncio import ensure_future
import functools
import os
import re
import shutil
from threading import Thread
from mongoengine import PULL
from mongomotor import Document, EmbeddedDocument
from mongomotor.fields import (StringField, IntField, ReferenceField,
                               DateTimeField, ListField, BooleanField,
                               EmbeddedDocumentField)
from toxicbuild.core import utils
from toxicbuild.master.build import BuildSet, Builder, BuildManager
from toxicbuild.master.exceptions import CloneException
from toxicbuild.master.plugins import MasterPlugin
from toxicbuild.master.pollers import Poller
from toxicbuild.master.signals import (build_started, build_finished,
                                       repo_status_changed, repo_added)
from toxicbuild.master.slave import Slave
from toxicbuild.master.utils import OwnedDocument


# The thing here is: When a repository poller is scheduled, I need to
# keep track of the hashes so I can remove it from the scheduler
# when needed.
# The format is {repourl: hash} for update_code
# and {repourl-start-pending: hash} for starting pending builds
_scheduler_hashes = {}


async def _update_code(repo_id):
    """Calls the update_code method of a repo. We must use this instead of
    use reload() in the update_code method because the reload() causes a
    massive memory leak."""

    repo = await Repository.get(id=repo_id)
    await repo.update_code()


class RepositoryBranch(EmbeddedDocument):
    name = StringField(required=True)
    notify_only_latest = BooleanField(default=False)

    def to_dict(self):
        return {'name': self.name,
                'notify_only_latest': self.notify_only_latest}


class Repository(OwnedDocument, utils.LoggerMixin):
    """Repository is where you store your code and where toxicbuild
    looks for incomming changes."""

    name = StringField(required=True, unique=True)
    url = StringField(required=True, unique=True)
    update_seconds = IntField(default=300, required=True)
    vcs_type = StringField(required=True, default='git')
    branches = ListField(EmbeddedDocumentField(RepositoryBranch))
    slaves = ListField(ReferenceField(Slave, reverse_delete_rule=PULL))
    clone_status = StringField(choices=('cloning', 'ready', 'clone-exception'),
                               default='cloning')
    plugins = ListField(EmbeddedDocumentField(MasterPlugin))
    # max number of builds in parallel that this repo exeutes
    # If None, there's no limit for parallel builds.
    parallel_builds = IntField()

    meta = {
        'ordering': ['name']
    }

    _plugins_instances = {}

    def __init__(self, *args, **kwargs):
        from toxicbuild.master import scheduler

        super(Repository, self).__init__(*args, **kwargs)
        self.scheduler = scheduler
        self._poller_instance = None
        self.build_manager = BuildManager(self)
        self._old_status = None

    async def to_dict(self, id_as_str=False):
        my_dict = {'id': self.id, 'name': self.name, 'url': self.url,
                   'update_seconds': self.update_seconds,
                   'vcs_type': self.vcs_type,
                   'branches': [b.to_dict() for b in self.branches],
                   'slaves': [s.to_dict(id_as_str)
                              for s in (await self.slaves)],
                   'plugins': [p.to_dict() for p in self.plugins],
                   'parallel_builds': self.parallel_builds,
                   'clone_status': self.clone_status}
        if id_as_str:
            my_dict['id'] = str(self.id)

        return my_dict

    @property
    def workdir(self):
        """ The directory where the source code of this repository is
        cloned into
        """

        workdir = re.sub(re.compile('http(s|)://'), '', self.url)
        workdir = workdir.replace('/', '-').replace('@', '-').replace(':', '')
        workdir = workdir.strip()
        return os.path.join('src', workdir)

    @property
    def poller(self):
        if self._poller_instance is None:
            vcs_type = self.vcs_type
            self._poller_instance = Poller(self, vcs_type, self.workdir)

        return self._poller_instance

    async def get_status(self):
        """Returns the status for the repository. The status is the
        status of the last buildset created for this repository that is
        not pending."""

        last_buildset = await BuildSet.objects(repository=self).order_by(
            '-created').first()

        clone_statuses = ['cloning', 'clone-exception']
        if not last_buildset and self.clone_status in clone_statuses:
            status = self.clone_status
        elif not last_buildset:
            status = 'ready'
        else:
            status = last_buildset.get_status()
            i = 1
            while status == BuildSet.PENDING:
                # we do not consider pending builds for the repo status
                start = i
                stop = start + 1
                last_buildset = BuildSet.objects(repository=self).order_by(
                    '-created')[start:stop]
                last_buildset = await last_buildset.first()

                if not last_buildset:
                    status = 'ready'
                    break

                status = last_buildset.get_status()
                i += 1
        return status

    @classmethod
    async def create(cls, name, url, owner, update_seconds, vcs_type,
                     slaves=None, branches=None, parallel_builds=None):
        """ Creates a new repository and schedule it.

        :param name: Repository name.
        :param url: Repository version control system url
        :param owner: :class:`~toxicbuild.master.users.User` or
          :class:`~toxicbuild.master.users.Organization` that owns the
          repository.
        :param update_seconds: How long we should wait until
          poll the changes again.
        :param vcs_type: Which type of version control system this
          repository uses.
        :param slaves: A list of slaves for this repository.
        :param branches: A list of branches config for this repository.
        :params parallel_builds: How many paralles builds this repository
          executes. If None, there is no limit."""

        slaves = slaves or []
        branches = branches or []

        repo = cls(url=url, update_seconds=update_seconds, vcs_type=vcs_type,
                   slaves=slaves, name=name, branches=branches,
                   parallel_builds=parallel_builds, owner=owner)
        await repo.save()
        repo_added.send(str(repo.id))
        repo.schedule()
        return repo

    async def remove(self):
        """ Removes all builds and builders and revisions related to the
        repository, removes the poller from the scheduler, removes the
        source code from the file system and then removes the repository.
        """

        builds = BuildSet.objects.filter(repository=self)
        await builds.delete()

        builders = Builder.objects.filter(repository=self)
        await builders.delete()

        revisions = RepositoryRevision.objects.filter(repository=self)
        await revisions.delete()

        try:
            sched_hash = _scheduler_hashes[self.url]
            self.scheduler.remove_by_hash(sched_hash)
            del _scheduler_hashes[self.url]

            pending_hash = _scheduler_hashes['{}-start-pending'.format(
                self.url)]
            self.scheduler.remove_by_hash(pending_hash)
            del _scheduler_hashes['{}-start-pending'.format(self.url)]
        except KeyError:  # pragma no cover
            # means the repository was not scheduled
            pass

        # removes the repository from the file system.
        Thread(target=shutil.rmtree, args=[self.workdir]).start()

        await self.delete()

    @classmethod
    async def get(cls, **kwargs):
        """Returns a repository instance

        :param kwargs: kwargs to match the repository."""
        repo = await cls.objects.get(**kwargs)
        return repo

    async def update_code(self):
        """Updates the repository's code. It is just a wrapper for
        self.poller.poll, so I can handle exceptions here."""

        with_clone = False
        try:
            with_clone = await self.poller.poll()
            clone_status = 'ready'
        except CloneException:
            with_clone = True
            clone_status = 'clone-exception'

        self.clone_status = clone_status
        await self.save()

        if with_clone:
            repo_status_changed.send(str(self.id), old_status='cloning',
                                     new_status=self.clone_status)

    def schedule(self):
        """Schedules all needed actions for a repository. The actions are:

        * Update source code using ``self.update_code``
        * Starts builds that are pending using
          ``self.build_manager.start_pending``.
        * Connects to ``build_started`` and ``build_finished`` signals
          to handle changing of status.
        * Runs the enabled plugins."""

        self.log('Scheduling {url}'.format(url=self.url))
        # we store this hashes so we can remove it from the scheduler when
        # we remove the repository.

        # adding update_code
        update_fn = functools.partial(_update_code, self.id)
        sched_hash = self.scheduler.add(update_fn, self.update_seconds)
        _scheduler_hashes[self.url] = sched_hash

        # adding start_pending
        start_pending_hash = self.scheduler.add(
            self.build_manager.start_pending, 120)
        _scheduler_hashes['{}-start-pending'.format(
            self.url)] = start_pending_hash

        # connecting to build signals
        build_started.connect(self._check_for_status_change)
        build_finished.connect(self._check_for_status_change)

        # starting plugins
        for plugin in self.plugins:
            self._run_plugin(plugin)

    @classmethod
    async def schedule_all(cls):
        """ Schedule all repositories. """

        repos = await cls.objects.all().to_list()
        for repo in repos:
            repo.schedule()

    async def add_slave(self, slave):
        """Adds a new slave to a repository.

        :param slave: A slave instance."""
        self.slaves
        slaves = await self.slaves
        slaves.append(slave)
        self.slaves = slaves
        await self.save()
        return slave

    async def remove_slave(self, slave):
        """Removes a slave from a repository.

        :param slave: A slave instance."""
        slaves = await self.slaves
        slaves.pop(slaves.index(slave))
        await self.update(set__slaves=slaves)
        return slave

    async def add_or_update_branch(self, branch_name,
                                   notify_only_latest=False):
        """Adds a new branch to this repository. If the branch
        already exists updates it with a new value.

        :param branch_name: The name of a branch
        :param notify_only_latest: If we should build only the most
          recent build of this branch"""

        # this is a shitty way of doing this. What is the
        # better way?
        def get_branch(branch_name):
            for b in self.branches:
                if b.name == branch_name:
                    return b

        branch = get_branch(branch_name)
        if branch:
            branch.notify_only_latest = notify_only_latest
        else:
            branch = RepositoryBranch(name=branch_name,
                                      notify_only_latest=notify_only_latest)
            self.branches.append(branch)

        await self.save()

    async def remove_branch(self, branch_name):
        """Removes a branch from this repository.

        :param branch_name: The branch name."""

        await self.update(pull__branches__name=branch_name)

    async def get_latest_revision_for_branch(self, branch):
        """ Returns the latest revision for a given branch

        :param branch: branch name
        """
        latest = RepositoryRevision.objects.filter(
            repository=self, branch=branch).order_by('-commit_date')

        latest = await latest.first()

        return latest

    async def get_latest_revisions(self):
        """ Returns the latest revision for all known branches
        """
        branches = await self.get_known_branches()
        revs = {}
        for branch in branches:
            rev = await self.get_latest_revision_for_branch(branch)
            revs[branch] = rev

        return revs

    async def get_known_branches(self):
        """ Returns the names for the branches that already have some
        revision here.
        """
        branches = await RepositoryRevision.objects.filter(
            repository=self).distinct('branch')

        return branches

    async def add_revision(self, branch, commit, commit_date, author, title):
        """ Adds a revision to the repository.

        :param commit: commit uuid
        :param branch: branch name
        :param commit_date: commit's date (on authors time)
        """
        revision = RepositoryRevision(repository=self, commit=commit,
                                      branch=branch, commit_date=commit_date,
                                      author=author, title=title)
        await revision.save()
        return revision

    def _run_plugin(self, plugin):
        key = '{}-plugin-{}'.format(self.url, plugin.name)
        type(self)._plugins_instances[key] = plugin
        ensure_future(plugin.run(self))

    def _stop_plugin(self, plugin):
        key = '{}-plugin-{}'.format(self.url, plugin.name)
        plugin = type(self)._plugins_instances[key]
        ensure_future(plugin.stop())
        del type(self)._plugins_instances[key]

    async def enable_plugin(self, plugin_name, **plugin_config):
        """Enables a plugin to this repository.

        :param plugin_name: The name of the plugin that is being enabled.
        :param plugin_config: A dictionary containing the plugin's
          configuration."""

        plugin_cls = MasterPlugin.get_plugin(name=plugin_name)
        plugin = plugin_cls(**plugin_config)
        self.plugins.append(plugin)
        await self.save()
        self._run_plugin(plugin)

    def _match_kw(self, plugin, **kwargs):
        """True if the plugin's attributes match the
        kwargs.

        :param plugin: A plugin instance.
        :param kwargs: kwargs to match the plugin"""

        for k, v in kwargs.items():
            try:
                attr = getattr(plugin, k)
            except AttributeError:
                return False
            else:
                if attr != v:
                    return False

        return True

    async def disable_plugin(self, **kwargs):
        """Disables a plugin to the repository.

        :param kwargs: kwargs to match the plugin."""
        matched = [p for p in self.plugins if self._match_kw(p, **kwargs)]
        for p in matched:
            self.plugins.remove(p)
            self._stop_plugin(p)
        await self.save()

    async def add_builds_for_slave(self, buildset, slave, builders=[]):
        """Adds a buildset to the build queue of a given slave
        for this repository.

        :param buildset: An instance of
          :class:`toxicbuild.master.build.BuildSet`.
        :param slave: An instance of :class:`toxicbuild.master.build.Slave`.
        """
        await self.build_manager.add_builds_for_slave(
            buildset, slave, builders=builders)

    async def _check_for_status_change(self, sender, build):
        """Called when a build is started or finished. If this event
        makes the repository change its status triggers a
        ``repo_status_changed`` signal.

        :param sender: The object that sent the signal
        :param build: The build that was started or finished"""

        status = await self.get_status()
        if status != self._old_status:
            repo_status_changed.send(str(self.id), old_status=self._old_status,
                                     new_status=status)
            self._old_status = status


class RepositoryRevision(Document):
    """A commit in the code tree."""

    repository = ReferenceField(Repository, required=True)
    commit = StringField(required=True)
    branch = StringField(required=True)
    author = StringField(required=True)
    title = StringField(required=True)
    commit_date = DateTimeField(required=True)

    @classmethod
    async def get(cls, **kwargs):
        ret = await cls.objects.get(**kwargs)
        return ret



# -*- coding: utf-8 -*-

# Copyright 2015-2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.


import asyncio
from asyncio import ensure_future
from collections import defaultdict, deque
from datetime import timedelta
import json
from uuid import uuid4
from mongoengine.queryset import queryset_manager
from mongomotor import Document, EmbeddedDocument
from mongomotor.fields import (StringField, ListField, EmbeddedDocumentField,
                               ReferenceField, DateTimeField, UUIDField,
                               IntField)
from toxicbuild.core.utils import (log, get_toxicbuildconf, now,
                                   list_builders_from_config, datetime2string,
                                   format_timedelta, LoggerMixin)
from toxicbuild.master.exceptions import DBError
from toxicbuild.master.signals import revision_added, build_added

# The statuses used in builds  ordered by priority.
ORDERED_STATUSES = ['running', 'exception', 'fail',
                    'warning', 'success', 'pending']


class SerializeMixin:

    """Simple mixin to serialization relatad stuff."""

    def to_dict(self, id_as_str=False):
        """ Transforms a Document into a dictionary.

        :param id_as_str: If true, transforms the id field into a string.
        """

        objdict = json.loads(super().to_json())
        objdict['id'] = str(self.id) if id_as_str else self.id
        return objdict

    async def async_to_json(self):
        """Async version of to_json. Expects a to_dict coroutine."""

        objdict = await self.to_dict(id_as_str=True)
        return json.dumps(objdict)


class Builder(SerializeMixin, Document):

    """ The entity responsible for executing the build steps.
    """

    name = StringField()
    repository = ReferenceField('toxicbuild.master.Repository')

    @classmethod
    async def create(cls, **kwargs):
        """Creates a new Builder.

        :param kwargs: kwargs passed to the builder constructor"""
        repo = cls(**kwargs)
        await repo.save()
        return repo

    @classmethod
    async def get(cls, **kwargs):
        """Returns a builder instance."""
        builder = await cls.objects.get(**kwargs)
        return builder

    @classmethod
    async def get_or_create(cls, **kwargs):
        """Returns a builder instance. If it does not exist, creates it.

        :param kwargs: kwargs to match the builder."""
        try:
            builder = await cls.get(**kwargs)
        except cls.DoesNotExist:
            builder = await cls.create(**kwargs)

        return builder

    async def get_status(self):
        """Returns the builder status."""

        try:
            qs = BuildSet.objects(builds__builder=self).order_by('-created')
            last_buildset = await qs[0]
        except (IndexError, AttributeError):
            status = 'idle'
        else:
            statuses = []
            for build in last_buildset.builds:
                builder = await build.builder
                if builder == self:
                    statuses.append(build.status)

            ordered_statuses = sorted(statuses,
                                      key=lambda i: ORDERED_STATUSES.index(i))
            status = ordered_statuses[0]

        return status

    async def to_dict(self, id_as_str=False):
        """Returns a dictionary for this builder.

        :param id_as_str: If true, the object id will be converted to string"""

        objdict = super().to_dict(id_as_str=id_as_str)
        objdict['status'] = await self.get_status()
        return objdict

    async def to_json(self):
        """Returns a json for this builder."""
        return (await self.async_to_json())


class BuildStep(EmbeddedDocument):

    """ A step for a build. This is the object that will store
    the step data. Who actually execute the steps is the slave.
    """

    STATUSES = ['running', 'fail', 'success', 'exception',
                'warning']

    uuid = UUIDField(required=True, default=lambda: uuid4())
    name = StringField(required=True)
    command = StringField(required=True)
    status = StringField(choices=STATUSES)
    output = StringField()
    started = DateTimeField(default=None)
    finished = DateTimeField(default=None)
    # the index of the step in the build.
    index = IntField(requred=True)
    total_time = IntField()

    def to_dict(self):
        objdict = json.loads(super().to_json())
        objdict['uuid'] = str(self.uuid)
        keys = objdict.keys()
        if 'started' not in keys:
            objdict['started'] = None
        else:
            objdict['started'] = datetime2string(self.started)

        if 'finished' not in keys:
            objdict['finished'] = None
        else:
            objdict['finished'] = datetime2string(self.finished)

        total = format_timedelta(timedelta(seconds=self.total_time)) \
            if self.total_time is not None else ''
        objdict['total_time'] = total

        return objdict

    def to_json(self):
        return json.dumps(self.to_dict())


class Build(EmbeddedDocument):

    """ A set of steps for a repository. This is the object that stores
    the build data. The build is carried by the slave.
    """

    PENDING = 'pending'
    STATUSES = BuildStep.STATUSES + [PENDING]

    uuid = UUIDField(required=True, default=lambda: uuid4())
    repository = ReferenceField('toxicbuild.master.Repository', required=True)
    slave = ReferenceField('Slave', required=True)
    branch = StringField(required=True)
    named_tree = StringField(required=True)
    started = DateTimeField()
    finished = DateTimeField()
    builder = ReferenceField(Builder, required=True)
    status = StringField(default=PENDING, choices=STATUSES)
    steps = ListField(EmbeddedDocumentField(BuildStep))
    total_time = IntField()

    def to_dict(self, id_as_str=False):
        steps = [s.to_dict() for s in self.steps]
        objdict = json.loads(super().to_json())
        objdict['builder']['id'] = objdict['builder']['$oid']
        objdict['uuid'] = str(self.uuid)
        objdict['steps'] = steps
        objdict['started'] = datetime2string(self.started) if self.started \
            else ''
        objdict['finished'] = datetime2string(self.finished) if self.finished \
            else ''
        if self.total_time is not None:
            td = timedelta(seconds=self.total_time)
            objdict['total_time'] = format_timedelta(td)
        else:
            objdict['total_time'] = ''
        return objdict

    def to_json(self):
        objdict = self.to_dict(id_as_str=True)
        return json.dumps(objdict)

    async def update(self):
        """Does an atomic update in this embedded document."""

        result = await BuildSet.objects(
            builds__uuid=self.uuid).update_one(set__builds__S=self)

        if not result:
            msg = 'This EmbeddedDocument was not save to database.'
            msg += ' You can\'t update it.'
            raise DBError(msg)

        return result

    async def get_buildset(self):
        """Returns the buildset that 'owns' this build."""
        buildset = await BuildSet.objects.get(builds__uuid=self.uuid)
        return buildset

    @property
    def output(self):
        """The build output. It is the commands + the output of the steps."""
        output = []
        for step in self.steps:
            output.append(step.command + '\n')
            output.append(step.output)

        return ''.join(output)


class BuildSet(SerializeMixin, Document):

    """A list of builds associated with a revision."""

    PENDING = Build.PENDING

    repository = ReferenceField('toxicbuild.master.Repository',
                                required=True)
    revision = ReferenceField('toxicbuild.master.RepositoryRevision',
                              required=True)
    commit = StringField(required=True)
    commit_date = DateTimeField(required=True)
    branch = StringField(required=True)
    author = StringField()
    title = StringField()
    builds = ListField(EmbeddedDocumentField(Build))
    # when this buildset was first created.
    created = DateTimeField(default=now)
    # when it actually started the builds
    started = DateTimeField()
    finished = DateTimeField()
    total_time = IntField()

    meta = {
        'indexes': [
            'repository'
        ]
    }

    @queryset_manager
    def objects(doc_cls, queryset):
        return queryset.order_by('created')

    @classmethod
    async def create(cls, repository, revision, save=True):
        """Creates a new buildset.

        :param repository: An instance of `toxicbuild.master.Repository`.
        :param revision: An instance of `toxicbuild.master.RepositoryRevision`.
        :param save: Indicates if the instance should be saved to database.
        """

        buildset = cls(repository=repository, revision=revision,
                       commit=revision.commit,
                       commit_date=revision.commit_date,
                       branch=revision.branch, author=revision.author,
                       title=revision.title)
        if save:
            await buildset.save()
        return buildset

    def to_dict(self, id_as_str=False):
        objdict = super().to_dict(id_as_str=id_as_str)
        objdict['commit_date'] = datetime2string(self.commit_date)
        objdict['created'] = datetime2string(self.created)
        objdict['started'] = datetime2string(self.started) if self.started \
            else ''
        objdict['finished'] = datetime2string(self.finished) if self.finished \
            else ''

        if self.total_time is not None:
            td = timedelta(seconds=self.total_time)
            objdict['total_time'] = format_timedelta(td)
        else:
            objdict['total_time'] = ''

        objdict['builds'] = []
        for b in self.builds:
            bdict = b.to_dict(id_as_str=id_as_str)
            objdict['builds'].append(bdict)
        return objdict

    def to_json(self):
        objdict = self.to_dict(id_as_str=True)
        return json.dumps(objdict)

    def get_status(self):
        build_statuses = set([b.status for b in self.builds])
        ordered_statuses = sorted(build_statuses,
                                  key=lambda i: ORDERED_STATUSES.index(i))
        return ordered_statuses[0]

    def get_pending_builds(self):
        return [b for b in self.builds if b.status == Build.PENDING]

    async def get_builds_for(self, builder=None, branch=None):

        async def match_builder(b):
            if not builder or (builder and (await b.builder).id == builder.id):
                return True
            return False

        def match_branch(b):
            if not branch or b.branch == branch:
                return True
            return False

        builds = []  # miss you py3.6
        for build in self.builds:
            if await match_builder(build) and match_branch(build):
                builds.append(build)
        return builds


class BuildManager(LoggerMixin):

    """ Controls which builds should be executed sequentially or
    in parallel.
    """
    # Note that this core reached its limit. It really needs a good
    # refactor, so we can implement builds that trigger other builds
    # in other slaves in a (almost) simple way.

    # each repository has its own key the default dict
    # each slave has its own queue
    _build_queues = defaultdict(lambda: defaultdict(deque))
    # to keep track of which slave is already working
    # on consume its queue
    _is_building = defaultdict(lambda: defaultdict(lambda: False))

    def __init__(self, repository):
        self.repository = repository
        self._is_getting_builders = False
        self._is_connected_to_signals = False
        self.connect2signals()

    @property
    def build_queues(self):
        return self._build_queues[self.repository.name]

    @property
    def is_building(self):
        return self._is_building[self.repository.name]

    async def add_builds(self, revisions):
        """ Adds the builds for a given revision in the build queue.

        :param revision: A list of
          :class:`toxicbuild.master.RepositoryRevision`
          instances for the build."""

        for revision in revisions:
            buildset = await BuildSet.create(repository=self.repository,
                                             revision=revision,
                                             save=False)

            slaves = await self.repository.slaves
            for slave in slaves:
                await self.add_builds_for_slave(buildset, slave)

    async def add_builds_for_slave(self, buildset, slave, builders=[]):
        """Adds builds for a given slave on a given buildset.

        :param buildset: An instance of :class:`toxicbuild.master.BuildSet`.
        :param slaves: An instance of :class:`toxicbuild.master.Slave`.
        :param builders: A list of :class:`toxicbuild.master.Builder`. If
          not builders all builders for this slave and revision will be used.
        """

        revision = await buildset.revision
        if not builders:
            builders = await self.get_builders(slave, revision)

        for builder in builders:
            build = Build(repository=self.repository, branch=revision.branch,
                          named_tree=revision.commit, slave=slave,
                          builder=builder)

            buildset.builds.append(build)
            await buildset.save()
            build_added.send(str(self.repository.id), build=build)

            self.log('build added for named_tree {} on branch {}'.format(
                revision.commit, revision.branch), level='debug')

        self.build_queues[slave.name].append(buildset)
        if not self.is_building[slave.name]:  # pragma: no branch
            ensure_future(self._execute_builds(slave))

    async def get_builders(self, slave, revision):
        """ Get builders for a given slave and revision.

        :param slave: A :class:`toxicbuild.master.slave.Slave` instance.
        :param revision: A
          :class:`toxicbuild.master.repository.RepositoryRevision`.
        """

        while self.repository.poller.is_polling() or self._is_getting_builders:
            await asyncio.sleep(1)

        self._is_getting_builders = True
        log('checkout on {} to {}'.format(
            self.repository.url, revision.commit), level='debug')
        try:
            await self.repository.poller.vcs.checkout(revision.commit)
            try:
                conf = get_toxicbuildconf(self.repository.poller.vcs.workdir)
                names = list_builders_from_config(conf, revision.branch, slave)
            except Exception as e:
                msg = 'Something wrong with your toxicbuild.conf. Original '
                msg += 'exception was:\n {}'.format(str(e))
                log(msg, level='warning')
                return []

            builders = []
            for name in names:
                builder = await Builder.get_or_create(
                    name=name, repository=self.repository)
                builders.append(builder)
        finally:
            self._is_getting_builders = False

        return builders

    async def start_pending(self):
        """Starts all pending buildsets that are not already scheduled for
        ``self.repository``."""

        buildsets = BuildSet.objects(builds__status=BuildSet.PENDING,
                                     repository=self.repository)
        buildsets = await buildsets.to_list()
        for buildset in buildsets:
            slaves = set()
            for b in buildset.builds:
                slave = await b.slave
                slaves.add(slave)

            for slave in slaves:
                # we schedule pending builds only if the slave is idle.
                if not self.build_queues[slave.name] and not \
                   self.is_building[slave.name]:

                    self.log('scheduling penging builds for {}'.format(
                        slave.name), level='debug')
                    self.log('schedule pending buildset {}'.format(str(
                        buildset.id)), level='debug')

                    self.build_queues[slave.name].append(buildset)
                    ensure_future(self._execute_builds(slave))

    def connect2signals(self):
        """ Connects the BuildManager to the revision_added signal."""

        revision_added.connect(self._revadded, sender=self.repository)
        self._is_connected_to_signals = True

    def disconnect_from_signals(self):
        revision_added.disconnect(self._revadded)
        self._is_connected_to_signals = False

    async def _revadded(self, sender, revisions):  # pragma no cover
        await self.add_builds(revisions)

    async def _set_started_for_buildset(self, buildset):
        if not buildset.started:
            buildset.started = now()
            await buildset.save()

    async def _set_finished_for_buildset(self, buildset):
        just_now = now()
        if not buildset.finished or buildset.finished < just_now:
            buildset.finished = just_now
            buildset.total_time = int((buildset.finished -
                                       buildset.started).seconds)
            await buildset.save()

    async def _execute_builds(self, slave):
        """ Execute the buildsets in the queue of a given slave.

        :param slave: A :class:`toxicbuild.master.slave.Slave` instance."""

        self.is_building[slave.name] = True
        self.log('executing builds for {}'.format(slave.name), level='debug')
        try:
            # here we take the buildsets that are in queue and send
            # each one of them to a slave execute the builds
            while True:
                try:
                    buildset = self.build_queues[slave.name].popleft()
                except IndexError:
                    break

                builds = []
                for build in buildset.builds:
                    build_slave = await build.slave
                    if slave == build_slave:
                        builds.append(build)
                if builds:
                    await self._set_started_for_buildset(buildset)
                    await self._execute_in_parallel(slave, builds)
                    await self._set_finished_for_buildset(buildset)
                    self.log('builds for {} finished'.format(slave.name),
                             level='debug')
        finally:
            self.is_building[slave.name] = False

    async def _execute_in_parallel(self, slave, builds):
        """Executes builds in parallel in a slave.

        :param slave: A :class:`toxicbuild.master.slave.Slave` instance.
        :param builds: A list of
          :class:`toxicbuild.master.build.Build` instances."""

        fs = []
        for chunk in self._get_builds_chunks(builds):
            chunk_fs = []
            for build in chunk:
                f = ensure_future(slave.build(build))
                chunk_fs.append(f)
                fs.append(f)
                await asyncio.wait(chunk_fs)

        return fs

    def _get_builds_chunks(self, builds):

        if not self.repository.parallel_builds:
            yield builds
            return

        for i in range(0, len(builds), self.repository.parallel_builds):
            yield builds[i:i + self.repository.parallel_builds]



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages


def get_version_from_file():
    # get version number from __init__ file
    # before module is installed

    fname = 'toxicbuild/__init__.py'
    with open(fname) as f:
        fcontent = f.readlines()
    version_line = [l for l in fcontent if 'VERSION' in l][0]
    return version_line.split('=')[1].strip().strip("'").strip('"')


def get_long_description_from_file():
    # content of README will be the long description

    fname = 'README'
    with open(fname) as f:
        fcontent = f.read()
    return fcontent


VERSION = get_version_from_file()

DESCRIPTION = """
Simple and flexible continuous integration tool.
""".strip()

LONG_DESCRIPTION = get_long_description_from_file()

setup(name='toxicbuild',
      version=VERSION,
      author='Juca Crispim',
      author_email='juca@poraodojuca.net',
      url='http://toxicbuild.poraodojuca.net',
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      packages=find_packages(exclude=['tests', 'tests.*']),
      license='GPL',
      include_package_data=True,
      install_requires=['tornado==4.4', 'mongomotor>=0.10.1',
                        'asyncblink>=0.1.1', 'mando>=0.3.2',
                        'pyrocumulus>=0.8rc0', 'pytz==2014.10',
                        'aiohttp>=1.0.5', 'aiosmtplib>=1.0.2'],
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: No Input/Output (Daemon)',
          'Environment :: Web Environment',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU General Public License (GPL)',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Topic :: Software Development :: Build Tools',
          'Topic :: Software Development :: Testing',
      ],
      entry_points={
          'console_scripts': ['toxicbuild=toxicbuild.script:main',
                              'toxicmaster=toxicbuild.master:main',
                              'toxicslave=toxicbuild.slave:main',
                              'toxicweb=toxicbuild.ui:main']
      },
      test_suite='tests',
      provides=['toxicbuild'],)



# -*- coding: utf-8 -*-

# Copyright 2015 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.


import os
from toxicbuild.core.exceptions import ConfigError
from toxicbuild.core.utils import load_module_from_file


class Settings:
    """ Simple interface to a settings file.
    """
    def __init__(self, envvar, default_filename):
        """:param envvar: Environment variable to look for settings file name
        :param default_filename: filename to use in case environment variable
        is not set.
        """
        self._envvar = envvar
        self._default_filename = default_filename
        self._settings_module = self._get_settings_module()

    @property
    def _filename(self):
        """ Returns the filename to use as settings module. Returns the filename
        on environment variable or ``self._default_filename``.
        """
        return self._get_settings_file_from_envvar() or self._default_filename

    def _get_settings_file_from_envvar(self):
        """
        Returns the value for the environment variable ``self._envvar``
        """
        settings_file = os.environ.get(self._envvar)

        return settings_file

    def _get_settings_module(self):
        """
        Returns the module to be used as the settings module.
        """
        module = load_module_from_file(self._filename)
        return module

    def __getattr__(self, attrname):
        try:
            attr = getattr(self._settings_module, attrname)
        except AttributeError:
            msg = 'Your settings {} file does not have an attribute {}.'.\
                  format(self._filename, attrname)
            raise ConfigError(msg)

        return attr

    def __setattr__(self, attrname, value):
        # haha!
        exceptions = ['_envvar', '_default_filename', '_settings_module']
        if attrname not in exceptions:
            raise AttributeError('you can\'t set here')

        super(Settings, self).__setattr__(attrname, value)



# -*- coding: utf-8 -*-

# Copyright 2015-2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

import asyncio
import os
from toxicbuild.core.exceptions import ConfigError
from toxicbuild.core.plugins import Plugin
from toxicbuild.core.utils import run_in_thread
from toxicbuild.slave import settings
from toxicbuild.slave.build import BuildStep


class SlavePlugin(Plugin):
    """This is a base slave plugin. Slave plugins may add steps to a build
    before and/or after the used defined steps. It may also set enivronment
    variables to be used in the tests."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # Your plugin must have an unique name
    name = 'BaseSlavePlugin'

    @property
    def data_dir(self):
        """The directory where the plugin store its data."""

        try:
            data_dir = settings.PLUGINS_DATA_DIR
        except ConfigError:
            data_dir = os.path.join('..', '.')

        return os.path.join(data_dir, self.name)

    async def create_data_dir(self):
        if not os.path.exists(self.data_dir):
            await run_in_thread(os.makedirs, self.data_dir)

    def get_steps_before(self):
        """Returns a list of steps to be executed before the steps provided
        by the user."""

        return []

    def get_steps_after(self):
        """Returns a list of steps to be executed after the steps provided
        by the user."""

        return []

    def get_env_vars(self):
        """ Returns a dictionary containing values for environment
        variables."""

        return {}


class PythonCreateVenvStep(BuildStep):

    """Step that checks if the venv already exists before
    executing the command."""

    def __init__(self, venv_dir, pyversion):
        self.venv_dir = venv_dir
        self.pyversion = pyversion
        name = 'Create virtualenv'
        command = 'virtualenv {} -p {}'.format(self.venv_dir,
                                               self.pyversion)
        super().__init__(name, command, stop_on_fail=True)

    @asyncio.coroutine
    def execute(self, cwd, **envvars):
        pyexec = os.path.join(self.venv_dir, os.path.join('bin', 'python'))
        if os.path.exists(os.path.join(cwd, pyexec)):
            step_info = {'status': 'success',
                         'output': 'venv exists. Skipping...'}
        else:
            step_info = yield from super().execute(cwd, **envvars)

        return step_info


class PythonVenvPlugin(SlavePlugin):
    name = 'python-venv'
    uses_data_dir = True

    def __init__(self, pyversion, requirements_file='requirements.txt',
                 remove_env=False):
        self.pyversion = pyversion
        self.requirements_file = requirements_file
        self.remove_env = remove_env
        self.venv_dir = os.path.join(
            self.data_dir, 'venv-{}'.format(
                self.pyversion.replace(os.sep, '')))

    def get_steps_before(self):
        create_env = PythonCreateVenvStep(self.venv_dir, self.pyversion)

        install_deps = BuildStep('install dependencies using pip',
                                 'pip install -r {}'.format(
                                     self.requirements_file),
                                 stop_on_fail=True)

        return [create_env, install_deps]

    def get_steps_after(self):
        steps = []
        if self.remove_env:
            steps.append(BuildStep('remove venv',
                                   'rm -rf {}'.format(self.venv_dir)))
        return steps

    def get_env_vars(self):
        return {'PATH': '{}/bin:PATH'.format(self.venv_dir)}


class AptUpdateStep(BuildStep):

    def __init__(self, timeout=600):
        cmd = 'sudo apt-get update'
        name = 'Updating apt packages list'
        super().__init__(name, cmd, stop_on_fail=True, timeout=timeout)


class AptInstallStep(BuildStep):

    def __init__(self, packages, timeout=600):
        packages = ' '.join(packages)
        cmd = ' '.join(['sudo apt-get install -y', packages])
        name = 'Installing packages with apt-get'
        super().__init__(name, cmd, stop_on_fail=True, timeout=timeout)


class AptInstallPlugin(SlavePlugin):

    """Installs packages using apt."""

    name = 'apt-install'

    def __init__(self, packages, timeout=600):
        """Initializes the plugin.
        :param packages: A list of packages names to be installed."""

        self.packages = packages

    def get_steps_before(self):
        update = AptUpdateStep()
        install = AptInstallStep(self.packages)
        return [update, install]

    def get_env_vars(self):
        return {'DEBIAN_FRONTEND': 'noninteractive'}



# -*- coding: utf-8 -*-


class BuilderNotFound(Exception):
    pass


class BadData(Exception):
    pass


class BadBuilderConfig(Exception):
    pass


class BusyRepository(Exception):
    pass


class NotConnected(Exception):
    pass



# -*- coding: utf-8 -*-

# Copyright 2015-2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.


import asyncio
from collections import OrderedDict
import json
from toxicbuild.ui import settings
from toxicbuild.core.utils import string2datetime
from toxicbuild.ui.client import get_hole_client
from toxicbuild.ui.utils import is_datetime


class BaseModel:
    # These references are fields that refer to other objects.
    # Note that this references are not always references on
    # database, they may be (and most are) embedded documents
    # that are simply treated as other objects.
    references = {}

    def __init__(self, requester, ordered_kwargs):
        # here is where we transform the dictonaries from the
        # master's response into objects that are references.
        # Note that we can't use **kwargs here because we want to
        # keep the order of the attrs.
        self.__ordered__ = [k for k in ordered_kwargs.keys()]

        for name, cls in self.references.items():
            if not isinstance(ordered_kwargs.get(name), (dict, cls)):
                ordered_kwargs[name] = [cls(requester, kw) if not
                                        isinstance(kw, cls)
                                        else kw
                                        for kw in ordered_kwargs.get(name, [])]
            else:
                obj = ordered_kwargs[name]
                ordered_kwargs[name] = cls(requester, obj) if not isinstance(
                    obj, cls) else obj

        for key, value in ordered_kwargs.items():
            if is_datetime(value):
                value = string2datetime(value)
            setattr(self, key, value)
            self.__ordered__.append(key)

        self.requester = requester

    def __eq__(self, other):
        return type(self) == type(other) and self.id == other.id

    def __hash__(self):
        return hash(self.id)

    @classmethod
    @asyncio.coroutine
    def get_client(cls, requester):
        """Returns a client connected to master."""

        host = settings.HOLE_HOST
        port = settings.HOLE_PORT
        client = yield from get_hole_client(requester, host, port)
        return client

    def to_dict(self):
        """Transforms a model into a dict."""

        attrs = [a for a in self.__ordered__ if not a.startswith('_')]

        d = OrderedDict()
        for attr in attrs:
            objattr = getattr(self, attr)
            is_ref = attr == 'references'
            if not (callable(objattr) and not is_ref):  # pragma no branch
                d[attr] = objattr

        return d

    def to_json(self):
        """Transforms a model into a json."""

        d = self.to_dict()
        return json.dumps(d)


class User(BaseModel):

    def __init__(self, requester, ordered_kwargs):
        if requester is None:
            requester = self
        super().__init__(requester, ordered_kwargs)

    @classmethod
    async def authenticate(cls, username_or_email, password):
        kw = {'username_or_email': username_or_email,
              'password': password}

        with (await cls.get_client(None)) as client:
            user_dict = await client.user_authenticate(**kw)
        user = cls(None, user_dict)
        return user

    @classmethod
    async def add(cls, requester, email, username, password,
                  allowed_actions):
        kw = {'username': username,
              'email': email,
              'password': password, 'allowed_actions': allowed_actions}

        with (await cls.get_client(requester)) as client:
            user_dict = await client.user_add(**kw)
        user = cls(None, user_dict)
        return user

    async def delete(self):
        kw = {'id': str(self.id)}
        with (await type(self).get_client(self.requester)) as client:
            resp = await client.user_remove(**kw)
        return resp


class Slave(BaseModel):

    @classmethod
    @asyncio.coroutine
    def add(cls, requester, name, host, port, token, owner):
        """Adds a new slave.

        :param name: Slave name.
        :param host: Slave host.
        :param port: Slave port.
        :param token: Authentication token.
        :param owner: The slave owner
        """

        kw = {'slave_name': name, 'slave_host': host,
              'slave_port': port, 'slave_token': token,
              'owner_id': str(owner.id)}

        with (yield from cls.get_client(requester)) as client:
            slave_dict = yield from client.slave_add(**kw)
        slave = cls(requester, slave_dict)
        return slave

    @classmethod
    @asyncio.coroutine
    def get(cls, requester, **kwargs):
        """Returns a slave.

        :param requester: The user who is requesting the operation.
        :param kwargs: kwargs to get the slave."""

        with (yield from cls.get_client(requester)) as client:
            slave_dict = yield from client.slave_get(**kwargs)
        slave = cls(requester, slave_dict)
        return slave

    @classmethod
    @asyncio.coroutine
    def list(cls, requester):
        """Lists all slaves.

        :param requester: The user who is requesting the operation."""

        with (yield from cls.get_client(requester)) as client:
            slaves = yield from client.slave_list()
        slave_list = [cls(requester, slave) for slave in slaves]
        return slave_list

    @asyncio.coroutine
    def delete(self):
        """Delete a slave."""

        with (yield from self.get_client(self.requester)) as client:
            resp = yield from client.slave_remove(slave_name=self.name)
        return resp

    @asyncio.coroutine
    def update(self, **kwargs):
        """Updates a slave"""
        with (yield from self.get_client(self.requester)) as client:
            resp = yield from client.slave_update(slave_name=self.name,
                                                  **kwargs)
        return resp


class Plugin(BaseModel):
    """A repository plugin."""

    @classmethod
    def list(cls, requester):
        """Lists all plugins available in the master.

        :param requester: The user who is requesting the operation."""

        with (yield from cls.get_client(requester)) as client:
            resp = yield from client.plugins_list()

        plugins = [cls(requester, p) for p in resp]
        return plugins

    @classmethod
    def get(cls, requester, name):
        """Return the schema for a specific plugin

        :param requester: The user who is requesting the operation.
        :param name: The name of the plugin"""

        with (yield from cls.get_client(requester)) as client:
            resp = yield from client.plugin_get(name=name)

        return cls(requester, resp)


class Repository(BaseModel):

    """Class representing a repository."""

    references = {'slaves': Slave,
                  'plugins': Plugin}

    @classmethod
    @asyncio.coroutine
    def add(cls, requester, name, url, owner, vcs_type, update_seconds=300,
            slaves=[], parallel_builds=None):
        """Adds a new repository.

        :param requester: The user who is requesting the operation.
        :param name: Repository's name.
        :param url: Repository's url.
        :param owner: The repository owner
        :param vcs_type: VCS type used on the repository.
        :param update_seconds: Interval to update the repository code.
        :param slaves: List with slave names for this reporitory.
        :params parallel_builds: How many paralles builds this repository
          executes. If None, there is no limit."""

        kw = {'repo_name': name, 'repo_url': url, 'vcs_type': vcs_type,
              'update_seconds': update_seconds,
              'parallel_builds': parallel_builds,
              'owner_id': str(owner.id)}

        kw.update({'slaves': slaves})
        with (yield from cls.get_client(requester)) as client:
            repo_dict = yield from client.repo_add(**kw)

        repo = cls(requester, repo_dict)
        return repo

    @classmethod
    @asyncio.coroutine
    def get(cls, requester, **kwargs):
        """Returns a repository.

        :param requester: The user who is requesting the operation.
        :param kwargs: kwargs to get the repository."""

        with (yield from cls.get_client(requester)) as client:
            repo_dict = yield from client.repo_get(**kwargs)
        repo = cls(requester, repo_dict)
        return repo

    @classmethod
    @asyncio.coroutine
    def list(cls, requester):
        """Lists all repositories.

        :param requester: The user who is requesting the operation."""

        with (yield from cls.get_client(requester)) as client:
            repos = yield from client.repo_list()
        repo_list = [cls(requester, repo) for repo in repos]
        return repo_list

    @asyncio.coroutine
    def delete(self):
        """Delete a repository."""

        with (yield from self.get_client(self.requester)) as client:
            resp = yield from client.repo_remove(repo_name=self.name)
        return resp

    @asyncio.coroutine
    def add_slave(self, slave):
        """Adds a slave to the repository.

        :param slave: A Slave instance."""

        with (yield from self.get_client(self.requester)) as client:
            resp = yield from client.repo_add_slave(repo_name=self.name,
                                                    slave_name=slave.name)
        return resp

    @asyncio.coroutine
    def remove_slave(self, slave):
        """Removes a slave from the repository.

        :param slave: A Slave instance.
        """

        with (yield from self.get_client(self.requester)) as client:
            resp = yield from client.repo_remove_slave(repo_name=self.name,
                                                       slave_name=slave.name)
        return resp

    @asyncio.coroutine
    def add_branch(self, branch_name, notify_only_latest):
        """Adds a branch config to a repositoiry.

        :param branch_name: The name of the branch.
        :param notify_only_latest: If we should create builds for all
          revisions or only for the lastest one."""

        with (yield from self.get_client(self.requester)) as client:
            resp = yield from client.repo_add_branch(
                repo_name=self.name, branch_name=branch_name,
                notify_only_latest=notify_only_latest)

        return resp

    @asyncio.coroutine
    def remove_branch(self, branch_name):
        """Removes a branch config from a repository.

        :param branch_name: The name of the branch."""
        with (yield from self.get_client(self.requester)) as client:
            resp = yield from client.repo_remove_branch(
                repo_name=self.name, branch_name=branch_name)

        return resp

    @asyncio.coroutine
    def update(self, **kwargs):
        """Updates a slave"""
        with (yield from self.get_client(self.requester)) as client:
            resp = yield from client.repo_update(repo_name=self.name,
                                                 **kwargs)
        return resp

    @asyncio.coroutine
    def start_build(self, branch, builder_name=None, named_tree=None,
                    slaves=[]):
        """Starts a (some) build(s) for a repository.

        :param branch: The name of the branch.
        :param builder_name: The name of the builder that will execute
          the build
        :param named_tree: The named_tree that will be builded. If no
          named_tree the last one will be used.
        :param slaves: A list with names of slaves that will execute
          the builds. If no slave is supplied all will be used."""

        with (yield from self.get_client(self.requester)) as client:
            resp = yield from client.repo_start_build(
                repo_name=self.name, branch=branch, builder_name=builder_name,
                named_tree=named_tree, slaves=slaves)
        return resp

    def to_dict(self):
        """Transforms a repository into a dictionary."""

        d = super().to_dict()
        d['slaves'] = [s.to_dict() for s in d['slaves']]
        d['plugins'] = [p.to_dict() for p in d['plugins']]
        return d

    def enable_plugin(self, plugin_name, **kwargs):
        """Enables a plugin to a repository.

        :param plugin_name: The plugin's name.
        :param kwargs: kwargs used to configure the plugin.
        """

        with (yield from self.get_client(self.requester)) as client:
            resp = yield from client.repo_enable_plugin(
                repo_name=self.name, plugin_name=plugin_name, **kwargs)

        return resp

    def disable_plugin(self, **kwargs):
        """Disables a plugin from a repository.

        :param kwargs: kwargs to match the plugin."""

        with (yield from self.get_client(self.requester)) as client:
            resp = yield from client.repo_disable_plugin(
                repo_name=self.name, **kwargs)

        return resp


class Builder(BaseModel):

    @classmethod
    @asyncio.coroutine
    def list(cls, requester, **kwargs):
        """Lists builders already used."""

        with (yield from cls.get_client(requester)) as client:
            builders = yield from client.builder_list(**kwargs)

        builders_list = [cls(requester, builder) for builder in builders]
        return builders_list


class Step(BaseModel):
    pass


class Build(BaseModel):
    references = {'steps': Step,
                  'builder': Builder}


class BuildSet(BaseModel):
    references = {'builds': Build}

    @classmethod
    @asyncio.coroutine
    def list(cls, requester, repo_name=None):
        """Lists buildsets. If ``repo_name`` only builds of this
        repsitory will be listed.

        :param repo_name: Name of a repository."""

        with (yield from cls.get_client(requester)) as client:
            buildsets = yield from client.buildset_list(repo_name=repo_name,
                                                        offset=10)

        buildset_list = [cls(requester, buildset) for buildset in buildsets]
        return buildset_list
